from typing import Any, Optional, Iterator, Union

from linkml.validator.plugins import ValidationPlugin
from linkml.validator.report import ValidationResult, Severity
from linkml.validator.validation_context import ValidationContext
from linkml_runtime import SchemaView


def _yield_quantity_value_objects(data: Any, path: Optional[list[Union[str, int]]] = None):
    """Recursively yield QuantityValue objects from data."""
    if path is None:
        path = []
    if isinstance(data, dict):
        if data.get("type") == "nmdc:QuantityValue":
            yield path, data
        else:
            # Recursively search nested dictionaries
            for key, value in data.items():
                yield from _yield_quantity_value_objects(value, path + [key])
    elif isinstance(data, list):
        # Handle lists of objects
        for i, item in enumerate(data):
            yield from _yield_quantity_value_objects(item, path + [i])


class NmdcSchemaValidationPlugin(ValidationPlugin):
    """A validation plugin which validates instances using NMDC-specific validation logic.

    This plugin is designed to be used as a part of LinkML's validation framework and in conjunction
    with the `JsonSchemaValidationPlugin` provided by LinkML. This plugin performs the following
    additional checks:

      1. Ensure that values for slots with range `QuantityValue` have a `unit` property that is in
         agreement with the `storage_unit` annotation on the slot
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._slot_storage_units_cache = {}

    def _slot_storage_units(self, schema_view: SchemaView, slot_name: str) -> Optional[list[str]]:
        """Get allowed storage_units for a slot."""
        if slot_name in self._slot_storage_units_cache:
            return self._slot_storage_units_cache[slot_name]

        slot = schema_view.get_slot(slot_name)
        if not slot or not slot.annotations:
            self._slot_storage_units_cache[slot_name] = None
            return None

        storage_units = None
        if "storage_units" in slot.annotations:
            annotation_obj = slot.annotations["storage_units"]
            if annotation_obj and hasattr(annotation_obj, "value"):
                # Split on pipes for multiple units
                storage_units = str(annotation_obj.value).split("|")

        self._slot_storage_units_cache[slot_name] = storage_units
        return storage_units

    def process(self, instance: Any, context: ValidationContext) -> Iterator[ValidationResult]:
        """Perform NMDC-specific schema validation on the provided instance

        :param instance: The instance to validate
        :param context: The validation context which provides a SchemaView artifact
        :return: Iterator over validation results
        :rtype: Iterator[ValidationResult]
        """
        for data_path, quantity_value in _yield_quantity_value_objects(instance):
            # Get the `has_unit` property from the QuantityValue instance. If it is missing, yield a
            # validation error and continue. This is slightly redundant with JSON Schema validation,
            # but it ensures that we do not attempt further validation on the instance.
            unit = quantity_value.get("has_unit")
            str_data_path = '/'.join(str(p) for p in data_path)
            if unit is None:
                yield ValidationResult(
                    type="nmdc-schema validation",
                    severity=Severity.ERROR,
                    instance=instance,
                    instantiates=context.target_class,
                    message=f"QuantityValue at /{str_data_path} is missing required 'has_unit' property",
                )
                continue

            # Find the slot name by looking for the last string in the data path. If one cannot be
            # found, skip further validation because we cannot determine the relevant slot.
            try:
                slot_name = next(key for key in reversed(data_path) if isinstance(key, str))
            except StopIteration:
                continue

            storage_units = self._slot_storage_units(context.schema_view, slot_name)
            if storage_units and unit not in storage_units:
                yield ValidationResult(
                    type="nmdc-schema validation",
                    severity=Severity.ERROR,
                    instance=instance,
                    instantiates=context.target_class,
                    message=(
                        f"QuantityValue at /{str_data_path} has unit '{unit}' which is not allowed for "
                        f"slot '{slot_name}' (allowed: {', '.join(storage_units)})"
                    )
                )



from linkml.validator import Validator

from nmdc_schema.migrators.migrator_base import MigratorBase
import sys
from pathlib import Path
from nmdc_schema.migrators.helpers import create_schema_view, get_classes_with_slots_by_range, \
    get_database_collections_for_class
from nmdc_schema import NmdcSchemaValidationPlugin

project_root = Path(__file__).parent.parent.parent
sys.path.insert(0, str(project_root))


class Migrator(MigratorBase):
    r"""A check-only migrator that raises an exception if any QuantityValue's has_unit slot
    is not valid against the slot's storage_unit or UnitEnum constraints."""

    _from_version = '11.13.0'
    _to_version = '11.14.0'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.schema_view = create_schema_view()
        self.validator = Validator(
            self.schema_view.schema,
            # This is intentionally *only* using the NMDC plugin, because this migrator is *only*
            # concerned with validating the QuantityValue-related constraints that the plugin
            # implements.
            validation_plugins=[
                NmdcSchemaValidationPlugin()
            ]
        )

    def upgrade(self, commit_changes: bool = False) -> None:
        r'''
        Migrates the database from conforming to the original schema, to conforming to the new schema.
        '''

        # Get the schema classes which have slots with range QuantityValue
        classes_with_qv_slots = get_classes_with_slots_by_range(self.schema_view, 'QuantityValue')

        # Get the Database collection names that can hold these classes
        eligible_collection_names = set()
        for class_name in classes_with_qv_slots.keys():
            collection_names = get_database_collections_for_class(self.schema_view, class_name)
            eligible_collection_names.update(collection_names)

        # Apply migrator through collections
        self.logger.info("Checking QuantityValue units against UnitEnum and storage_units constraints")
        for collection_name in eligible_collection_names:
            self.logger.info(f"  Checking collection '{collection_name}'")
            self.adapter.do_for_each_document(collection_name, self.confirm_units_fit_unitenum_and_storage_units)

    def confirm_units_fit_unitenum_and_storage_units(self, document: dict) -> None:
        r'''
        Raise an exception if the QuantityValue's has_unit slot is not valid against slot's storage_unit or UnitEnum constraints.

        >>> m = Migrator()

        # Test: valid QuantityValue with proper units in biosample
        >>> valid_biosample = {
        ...     "id": "test1",
        ...     "type": "nmdc:Biosample",
        ...     "bulk_elect_conductivity": {
        ...         "type": "nmdc:QuantityValue",
        ...         "has_unit": "mS/cm",
        ...         "has_numeric_value": 25.0
        ...     }
        ... }
        >>> m.confirm_units_fit_unitenum_and_storage_units(valid_biosample)  # Should not raise

        # Test: unit not allowed for bulk_elect_conductivity's storage_units
        >>> invalid_biosample_storage = {
        ...     "id": "test2",
        ...     "type": "nmdc:Biosample",
        ...     "bulk_elect_conductivity": {
        ...         "type": "nmdc:QuantityValue",
        ...         "has_unit": "Cel",
        ...         "has_numeric_value": 25.0
        ...     }
        ... }
        >>> m.confirm_units_fit_unitenum_and_storage_units(invalid_biosample_storage)
        Traceback (most recent call last):
            ...
        ValueError: In test2:
          QuantityValue at /bulk_elect_conductivity has unit 'Cel' which is not allowed for slot 'bulk_elect_conductivity' (allowed: mS/cm)

        # Test: unit not allowed for substances_volume's storage_units
        >>> invalid_chem_storage = {
        ...     "id": "test3",
        ...     "type": "nmdc:ChemicalConversionProcess",
        ...     "substances_volume": {
        ...         "type": "nmdc:QuantityValue",
        ...         "has_unit": "J/K",  # Wrong unit type for substances_volume
        ...         "has_numeric_value": 25.0
        ...     }
        ... }
        >>> m.confirm_units_fit_unitenum_and_storage_units(invalid_chem_storage)
        Traceback (most recent call last):
            ...
        ValueError: In test3:
          QuantityValue at /substances_volume has unit 'J/K' which is not allowed for slot 'substances_volume' (allowed: mL)
        '''

        document_id = document.get('id', '<unknown id>')
        document_type = document.get('type')
        if not document_type:
            raise ValueError(f"Unable to infer target_class for document with id '{document_id}'")
        target_class = document_type.replace('nmdc:', '')
        report = self.validator.validate(document, target_class)
        if report.results:
            raise ValueError(f"In {document_id}:\n" + "\n".join("  " + result.message for result in report.results))



from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    r"""
    Migrates a database between two schemas.
    """

    _from_version = "11.13.0.part_2"
    _to_version = "11.13.0.part_3"

    def upgrade(self, commit_changes: bool = False) -> None:
        r"""
        Migrates the database from conforming to the original schema, to conforming to the new schema.
        """
        self.adapter.do_for_each_document("biosample_set", self.confirm_igsn_prefix)

    def confirm_igsn_prefix(self, biosample: dict) -> None:
        r"""
        If a biosample  has a igsn_biosample_identifiers prefix of 'IGSN' raise an exception.

        >>> m = Migrator()

        # Test: igsn_biosample_identifiers matching "IGSN:"
        >>> m.confirm_igsn_prefix({"id": 1, "type": "nmdc:Biosample", "igsn_biosample_identifiers": ["IGSN:AU124"]})
        Traceback (most recent call last):
            ...
        ValueError: Biosample 1 'igsn_biosample_identifiers' list includes: IGSN:AU124

        # Test: igsn_biosample_identifiers with another non-confirming prefix
        >>> m.confirm_igsn_prefix({"id": 1, "type": "nmdc:Biosample", "igsn_biosample_identifiers": ["isng:AU124"]})
        Traceback (most recent call last):
            ...
        ValueError: Biosample 1 'igsn_biosample_identifiers' list includes: isng:AU124

        # Test: valid Biosample
        >>> m.confirm_igsn_prefix({"id": 3, "type": "nmdc:Biosample", "igsn_biosample_identifiers": ["igsn:AU124"]}) is None
        True
        """
        prefix = "igsn:"
        igsn_values = biosample.get("igsn_biosample_identifiers", [])
        biosample_id = biosample.get("id")
        for record in igsn_values:
            if not record.startswith(prefix):
                raise ValueError(
                    f"Biosample {biosample_id} 'igsn_biosample_identifiers' list includes: {record}"
                )



from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    r"""
    Migrates a database between two schemas.
    """

    _from_version = "PR176"
    _to_version = "PR104"

    def upgrade(self) -> None:
        r"""
        Deletes specific fields from documents (whose `type` is `nmdc:MassSpectrometry`) residing in
        the `data_generation_set` collection.
        """

        self.adapter.process_each_document(
            collection_name="data_generation_set",
            pipeline=[self.delete_obsolete_fields],
        )

    @staticmethod
    def delete_obsolete_fields(data_generation: dict) -> dict:
        r"""
        Deletes specific fields from a document representing a `DataGeneration` instance.

        If the document's `type` is not `nmdc:MassSpectrometry`, this function will not modify the document.
        If a target field exists and is empty, this function will delete the field. On the other hand, if a target
        field exists and is not empty, this function will not delete the field and will, instead, raise an exception.

        >>> m = Migrator()
        >>> m.delete_obsolete_fields({'id': 123, 'type': 'nmdc:MassSpectrometry'})  # no obsolete fields exist
        {'id': 123, 'type': 'nmdc:MassSpectrometry'}
        >>> m.delete_obsolete_fields({'id': 123, 'type': 'nmdc:NucleotideSequencing',
        ...                           'ncbi_project_name': ''})  # type is not relevant
        {'id': 123, 'type': 'nmdc:NucleotideSequencing', 'ncbi_project_name': ''}
        >>> m.delete_obsolete_fields({'id': 123, 'type': 'nmdc:MassSpectrometry',
        ...                           'ncbi_project_name': ''})  # single-valued, is empty
        {'id': 123, 'type': 'nmdc:MassSpectrometry'}
        >>> m.delete_obsolete_fields({'id': 123, 'type': 'nmdc:MassSpectrometry',
        ...                           'ncbi_project_name': None})
        {'id': 123, 'type': 'nmdc:MassSpectrometry'}
        >>> m.delete_obsolete_fields({'id': 123, 'type': 'nmdc:MassSpectrometry',
        ...                           'gold_sequencing_project_identifiers': []})  # multivalued, is empty
        {'id': 123, 'type': 'nmdc:MassSpectrometry'}
        >>> m.delete_obsolete_fields({'id': 123, 'type': 'nmdc:MassSpectrometry',
        ...                           'gold_sequencing_project_identifiers': None})
        {'id': 123, 'type': 'nmdc:MassSpectrometry'}
        >>> m.delete_obsolete_fields({'id': 123, 'type': 'nmdc:MassSpectrometry',
        ...                           'ncbi_project_name': 'a'})  # single-valued, is not empty
        Traceback (most recent call last):
        ...
        ValueError: Field "ncbi_project_name" in document "123" is not empty (contains value "a").
        >>> m.delete_obsolete_fields({'id': 123, 'type': 'nmdc:MassSpectrometry',
        ...                           'gold_sequencing_project_identifiers': ['a']})  # multivalued, is not empty
        Traceback (most recent call last):
        ...
        ValueError: Field "gold_sequencing_project_identifiers" in document "123" is not empty (contains value "['a']").
        >>> m.delete_obsolete_fields({'id': 123, 'type': 'nmdc:MassSpectrometry',
        ...                           'ncbi_project_name': None,
        ...                           'target_gene': '',
        ...                           'target_subfragment': '',
        ...                           'gold_sequencing_project_identifiers': None,
        ...                           'insdc_bioproject_identifiers': [],
        ...                           'insdc_experiment_identifiers': []})  # multiple obsolete fields
        {'id': 123, 'type': 'nmdc:MassSpectrometry'}
        """

        # If this document's `type` is not "nmdc:MassSpectrometry", leave the document as-is.
        if data_generation["type"] != "nmdc:MassSpectrometry":
            return data_generation

        # List the names of the single-valued fields we want to delete.
        # Note: Their emptiness is represented by `None` or `""`.
        single_valued_fields_to_delete = [
            "ncbi_project_name",
            "target_gene",
            "target_subfragment",
        ]

        # List the names of the multivalued fields we want to delete.
        # Note: Their emptiness is represented by `None` or `[]`.
        multi_valued_fields_to_delete = [
            "gold_sequencing_project_identifiers",
            "insdc_bioproject_identifiers",
            "insdc_experiment_identifiers",
        ]

        # Combine the two lists.
        field_names = single_valued_fields_to_delete + multi_valued_fields_to_delete

        for field_name in field_names:
            # Check whether the field exists.
            if field_name in data_generation:
                value = data_generation[field_name]

                # Check whether the field is empty (using emptiness criteria appropriate for the field).
                if (
                    field_name in single_valued_fields_to_delete and value in [None, ""]
                ) or (
                    field_name in multi_valued_fields_to_delete and value in [None, []]
                ):
                    del data_generation[field_name]
                else:
                    document_id = data_generation["id"]
                    raise ValueError(
                        f'Field "{field_name}" in document "{document_id}" is not empty (contains value "{value}").'
                    )

        return data_generation



from nmdc_schema.migrators.migrator_base import MigratorBase
from nmdc_schema.migrators.helpers import load_yaml_asset

class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "11.8.0.part_2"
    _to_version = "11.8.0.part_3"

    def upgrade(self) -> None:
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""
        self.adapter.do_for_each_document("data_generation_set", self.validate_mass_spec_slots)

    def validate_mass_spec_slots(self, data_generation_record: dict) -> None:
        r"""
        If the data object is of type `nmdc:MassSpectrometry` and either (a) its `eluent_introduction_category` field is missing or consists of `None`, or (b) its `has_mass_spectrometry_configuration` field is missing or consists of `None`, raise a `ValueError`.

        >>> m = Migrator()
        >>> m.validate_mass_spec_slots({"id": 123, "type": "nmdc:MassSpectrometry", "eluent_introduction_category": "gas_chromatography", "has_mass_spectrometry_configuration": "nmdc:mscon-123"})
        >>> m.validate_mass_spec_slots({"id": 123, "type": "nmdc:MassSpectrometry", "eluent_introduction_category": "gas_chromatography"})
        Traceback (most recent call last):
        ...
        ValueError: `has_mass_spectrometry_configuration` is required and is not present in the data object 123
        >>> m.validate_mass_spec_slots({"id": 123, "type": "nmdc:MassSpectrometry", "has_mass_spectrometry_configuration": "nmdc:mscon-123"})
        Traceback (most recent call last):
        ... 
        ValueError: `eluent_introduction_category` is required and is not present in the data object 123
        """
        # get the eluent_introduction_category and has_mass_spectrometry_configuration fields
        if data_generation_record.get("type") == "nmdc:MassSpectrometry":
            eluent_introduction_category = data_generation_record.get("eluent_introduction_category")
            if eluent_introduction_category is None:
                raise ValueError(f"`eluent_introduction_category` is required and is not present in the data object {data_generation_record.get('id')}")
            
            has_mass_spec_config = data_generation_record.get("has_mass_spectrometry_configuration")
            if has_mass_spec_config is None:
                raise ValueError(f"`has_mass_spectrometry_configuration` is required and is not present in the data object {data_generation_record.get('id')}")



r"""
This module contains helper functions related to schema element identifiers.
"""

import re
from typing import List, Optional


def get_compatible_typecodes(slot_pattern: str) -> List[str]:
    r"""
    Returns the typecodes, if any, that the schema says values of the specified `id` field can contain.

    :param slot_pattern: The value of the `pattern` property of the `id` slot definition

    Note: We currently use the `pattern` property of the `id` slot for two different things: (a) to specify which
          typecode we want ID generators (e.g., instances of the NMDC ID Minter) to use when generating new identifiers
          (specifically, we want them to use only the _first_ typecode that occurs in the pattern), and (b) to specify
          which typecodes we want ID validators (e.g., instance of the NMDC Runtime) to allow identifiers of instances
          of that class to contain (they can contain _any_ one of the typecodes that occurs in the pattern). This
          function returns the list for "use (b)." Eventually, we may specify the above two things using separate
          schema elements.

    >>> get_compatible_typecodes("(nmdc):foo-...")
    []
    >>> get_compatible_typecodes("^(nmdc):foo-...")
    ['foo']
    >>> get_compatible_typecodes("^(nmdc):(foo|bar)-...")
    ['foo', 'bar']
    >>> get_compatible_typecodes("^(nmdc):(foo|bar|baz)-...")
    ['foo', 'bar', 'baz']
    """
    typecodes: List[str] = []

    # Make a regular expression that can be used to extract the typecode portion of a slot pattern.
    #
    # Examples:
    # - "^(nmdc):foo-..."           → "foo"
    # - "^(nmdc):(foo|bar|baz)-..." → "(foo|bar|baz)"
    #
    # Note: The `+?` is a lazy matcher (opposite of greedy). If we did a greedy match here, the capture group would
    #       consume all but the final hyphen in the slot pattern. By doing a lazy match, we prevent the capture
    #       group from consuming any hyphens at all.
    #
    id_pattern = re.compile(r"^\^\(nmdc\):(.+?)-")

    # Use the regular expression to extract the typecode portion of the specified slot pattern.
    match_obj = id_pattern.match(slot_pattern)
    if match_obj is not None:
        typecode_portion = match_obj.group(1)

        # Extract the typecode from a single-typecode pattern, or extract all typecodes from a multi-typecode pattern.
        if "(" not in typecode_portion:
            # Example: "foo" → ["foo"]
            typecodes = [typecode_portion]
        else:
            # Example: "(foo|bar|baz)" → ["foo", "bar", "baz"]
            typecodes = typecode_portion.lstrip("(").rstrip(")").split("|")

    return typecodes


def get_typecode_for_future_ids(slot_pattern: str) -> Optional[str]:
    r"""
    Returns the typecode, if any, that schema authors want future values of the specified `id` field to contain.

    :param slot_pattern: The value of the `pattern` property of the `id` slot definition

    >>> get_typecode_for_future_ids("(nmdc):foo-...") is None
    True
    >>> get_typecode_for_future_ids("^(nmdc):foo-...")
    'foo'
    >>> get_typecode_for_future_ids("^(nmdc):(foo|bar)-...")
    'foo'
    >>> get_typecode_for_future_ids("^(nmdc):(foo|bar|baz)-...")
    'foo'
    """
    compatible_typecodes = get_compatible_typecodes(slot_pattern)
    return compatible_typecodes[0] if len(compatible_typecodes) > 0 else None



from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "8.0"
    _to_version = "8.1"

    def upgrade(self):
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""

        self.adapter.process_each_document("study_set", [self.force_research_study_study_category])

    def force_research_study_study_category(self, study: dict) -> dict:
        r"""
        If the study lacks a field named `study_category`, creates it and assigns it the value "research_study".

        >>> m = Migrator()
        >>> m.force_research_study_study_category({'id': 123})  # field doesn't exist yet
        {'id': 123, 'study_category': 'research_study'}
        >>> m.force_research_study_study_category({'id': 123, 'study_category': 'preserve_me'})  # field already exists
        {'id': 123, 'study_category': 'preserve_me'}
        """

        if "study_category" not in study:
            self.logger.info(f"Forcing 'study_category: research_study' on {study['id']}")
            study["study_category"] = "research_study"
        return study



from typing import List, Type

from nmdc_schema.migrators.migrator_base import MigratorBase
from nmdc_schema.migrators.partials.migrator_from_11_8_0_to_11_9_0 import migrator_from_11_8_0_to_11_9_0_part_1
from nmdc_schema.migrators.partials.migrator_from_11_8_0_to_11_9_0 import migrator_from_11_8_0_to_11_9_0_part_2


def get_migrator_classes() -> List[Type[MigratorBase]]:
    r"""
    Returns a list of migrator classes in the order in which they (i.e. their `upgrade` methods)
    were designed to be run.

    >>> migrator_classes = get_migrator_classes()
    >>> type(migrator_classes) is list and len(migrator_classes) > 0  # the function returns a list
    True
    >>> from inspect import isclass
    >>> all(isclass(c) for c in migrator_classes)  # each list item is a class
    True
    >>> all(callable(getattr(c, "upgrade")) for c in migrator_classes)  # each class has an `upgrade` method
    True
    """

    return [
        migrator_from_11_8_0_to_11_9_0_part_1.Migrator,
        migrator_from_11_8_0_to_11_9_0_part_2.Migrator,
    ]



from nmdc_schema.migrators.migrator_base import MigratorBase

class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "11.8.0.part_3"
    _to_version = "11.8.0.part_4"

    def upgrade(self) -> None:
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""
        self.adapter.do_for_each_document("configuration_set", self.validate_mass_spec_config_slots)
        self.adapter.do_for_each_document("configuration_set", self.validate_chrom_config_slots)

    def validate_mass_spec_config_slots(self, configuration_record: dict) -> None:
        r"""
        If the configuration record is of type MassSpectrometryConfiguration, does not have mass_spectrometry_acquisition_strategy, resolution_categories, mass_analyzers, ionization_source, mass_spectrum_collection_modes, and polarity_mode OR any of those keys does not have a value, raise a ValueError.

        >>> m = Migrator()
        >>> m.validate_mass_spec_config_slots({"id": 123, "type": "nmdc:MassSpectrometryConfiguration", "mass_spectrometry_acquisition_strategy": "data_independent_acquisition", "resolution_categories": "high", "mass_analyzers": "ion_trap", "ionization_source": "electron_ionization", "mass_spectrum_collection_modes": "centroid", "polarity_mode": "positive"})
        >>> m.validate_mass_spec_config_slots({"id": 123, "type": "nmdc:MassSpectrometryConfiguration", "mass_spectrometry_acquisition_strategy": "data_independent_acquisition", "resolution_categories": "high", "mass_analyzers": "ion_trap", "ionization_source": "electron_ionization", "mass_spectrum_collection_modes": "centroid"})
        Traceback (most recent call last):
        ...
        ValueError: `polarity_mode` is required and is not present in the configuration record 123
        >>> m.validate_mass_spec_config_slots({"id": 123, "type": "nmdc:MassSpectrometryConfiguration", "mass_spectrometry_acquisition_strategy": "data_independent_acquisition", "resolution_categories": "high", "mass_analyzers": "ion_trap", "ionization_source": "electron_ionization", "polarity_mode": "positive"})
        Traceback (most recent call last):
        ... 
        ValueError: `mass_spectrum_collection_modes` is required and is not present in the configuration record 123
        >>> m.validate_mass_spec_config_slots({"id": 123, "type": "nmdc:MassSpectrometryConfiguration", "mass_spectrometry_acquisition_strategy": "data_independent_acquisition", "resolution_categories": "high", "mass_analyzers": "ion_trap", "mass_spectrum_collection_modes": "centroid", "polarity_mode": "positive"})
        Traceback (most recent call last):
        ...
        ValueError: `ionization_source` is required and is not present in the configuration record 123
        >>> m.validate_mass_spec_config_slots({"id": 123, "type": "nmdc:MassSpectrometryConfiguration", "mass_spectrometry_acquisition_strategy": "data_independent_acquisition", "resolution_categories": "high", "mass_spectrum_collection_modes": "centroid", "polarity_mode": "positive"})
        Traceback (most recent call last):
        ...
        ValueError: `mass_analyzers` is required and is not present in the configuration record 123
        >>> m.validate_mass_spec_config_slots({"id": 123, "type": "nmdc:MassSpectrometryConfiguration", "mass_spectrometry_acquisition_strategy": "data_independent_acquisition", "mass_analyzers": "ion_trap", "ionization_source": "electron_ionization", "mass_spectrum_collection_modes": "centroid", "polarity_mode": "positive"})
        Traceback (most recent call last):
        ...
        ValueError: `resolution_categories` is required and is not present in the configuration record 123
        >>> m.validate_mass_spec_config_slots({"id": 123, "type": "nmdc:MassSpectrometryConfiguration", "resolution_categories": "high", "mass_analyzers": "ion_trap", "ionization_source": "electron_ionization", "mass_spectrum_collection_modes": "centroid", "polarity_mode": "positive"})
        Traceback (most recent call last):
        ...
        ValueError: `mass_spectrometry_acquisition_strategy` is required and is not present in the configuration record 123
        """
        config_record_id = configuration_record.get("id")
        # get the slots that are required for mass spectrometry configuration
        if configuration_record.get("type") == "nmdc:MassSpectrometryConfiguration":
            mass_spectrometry_acquisition_strategy = configuration_record.get("mass_spectrometry_acquisition_strategy")
            if mass_spectrometry_acquisition_strategy is None:
                raise ValueError(f"`mass_spectrometry_acquisition_strategy` is required and is not present in the configuration record {config_record_id}")
            
            resolution_categories = configuration_record.get("resolution_categories")
            if resolution_categories is None:
                raise ValueError(f"`resolution_categories` is required and is not present in the configuration record {config_record_id}")
            
            mass_analyzers = configuration_record.get("mass_analyzers")
            if mass_analyzers is None:
                raise ValueError(f"`mass_analyzers` is required and is not present in the configuration record {config_record_id}")
            
            ionization_source = configuration_record.get("ionization_source")
            if ionization_source is None:
                raise ValueError(f"`ionization_source` is required and is not present in the configuration record {config_record_id}")
            
            mass_spectrum_collection_modes = configuration_record.get("mass_spectrum_collection_modes")
            if mass_spectrum_collection_modes is None:
                raise ValueError(f"`mass_spectrum_collection_modes` is required and is not present in the configuration record {config_record_id}")
            
            polarity_mode = configuration_record.get("polarity_mode")
            if polarity_mode is None:
                raise ValueError(f"`polarity_mode` is required and is not present in the configuration record {config_record_id}")
    
    def validate_chrom_config_slots(self, configuration_record:dict) -> None:
        r"""
        If the configuration record is of type `ChromatographyConfiguration` and either (a) its `chromatographic_category` field is missing or consists of `None`, or (b) its `stationary_phase` field is missing or consists of `None`, raise a `ValueError`.

        >>> m = Migrator()
        >>> m.validate_chrom_config_slots({"id": 123, "type": "nmdc:ChromatographyConfiguration", "chromatographic_category": "gas_chromatography", "stationary_phase": "C18"})
        >>> m.validate_chrom_config_slots({"id": 123, "type": "nmdc:ChromatographyConfiguration", "chromatographic_category": "gas_chromatography"})
        Traceback (most recent call last):
        ...
        ValueError: `stationary_phase` is required and is not present in the configuration record 123
        >>> m.validate_chrom_config_slots({"id": 123, "type": "nmdc:ChromatographyConfiguration", "stationary_phase": "C18"})
        Traceback (most recent call last):
        ...
        ValueError: `chromatographic_category` is required and is not present in the configuration record 123
        """
        config_record_id = configuration_record.get("id")
        if configuration_record.get("type") == "nmdc:ChromatographyConfiguration":
            chromatographic_category = configuration_record.get("chromatographic_category")
            if chromatographic_category is None:
                raise ValueError(f"`chromatographic_category` is required and is not present in the configuration record {config_record_id}")
            
            stationary_phase = configuration_record.get("stationary_phase")
            if stationary_phase is None:
                raise ValueError(f"`stationary_phase` is required and is not present in the configuration record {config_record_id}")
            



from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "11.10.0"
    _to_version = "11.11.0.part_1"

    def upgrade(self, commit_changes: bool = True) -> None:
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""
        self.adapter.do_for_each_document("biosample_set", self.check_for_fields)

    def check_for_fields(self, biosample: dict) -> None:
        r"""
        Check each biosample record to ensure none of the removed slots are being used. 
        List of the slots that were removed:
        - dna_absorb1
        - dna_absorb2
        - dna_collect_site
        - dna_concentration
        - dna_cont_type
        - dna_cont_well
        - dna_container_id
        - dna_dnase
        - dna_isolate_meth
        - dna_organisms
        - dna_project_contact
        - dna_samp_id
        - dna_sample_format
        - dna_sample_name
        - dna_seq_project
        - dna_seq_project_pi
        - dna_seq_project_name
        - dna_volume
        - proposal_dna
        - dnase_rna
        - proposal_rna
        - rna_absorb1
        - rna_absorb2
        - rna_collect_site
        - rna_concentration
        - rna_cont_type
        - rna_cont_well
        - rna_container_id
        - rna_isolate_meth
        - rna_organisms
        - rna_project_contact
        - rna_samp_id
        - rna_sample_format
        - rna_sample_name
        - rna_seq_project
        - rna_seq_project_pi
        - rna_seq_project_name
        - rna_volume
        - collection_date_inc

        >>> m = Migrator()
        >>> m.check_for_fields({"id":123, "type": "nmdc:Biosample", "dna_absorb1": "value"})
        Traceback (most recent call last):
        ...
        Exception: Field `dna_absorb1` present in biosample 123.
        >>> m.check_for_fields({"id":123, "type": "nmdc:Biosample", "dna_absorb1": ""})
        Traceback (most recent call last):
        ...
        Exception: Field `dna_absorb1` present in biosample 123.
        >>> m.check_for_fields({"id":123, "type": "nmdc:Biosample"})
        """
        id = biosample.get("id")
        removed_slots = [
            "dna_absorb1",
            "dna_absorb2",
            "dna_collect_site",
            "dna_concentration",
            "dna_cont_type",
            "dna_cont_well",
            "dna_container_id",
            "dna_dnase",
            "dna_isolate_meth",
            "dna_organisms",
            "dna_project_contact",
            "dna_samp_id",
            "dna_sample_format",
            "dna_sample_name",
            "dna_seq_project",
            "dna_seq_project_pi",
            "dna_seq_project_name",
            "dna_volume",
            "proposal_dna",
            "dnase_rna",
            "proposal_rna",
            "rna_absorb1",
            "rna_absorb2",
            "rna_collect_site",
            "rna_concentration",
            "rna_cont_type",
            "rna_cont_well",
            "rna_container_id",
            "rna_isolate_meth",
            "rna_organisms",
            "rna_project_contact",
            "rna_samp_id",
            "rna_sample_format",
            "rna_sample_name",
            "rna_seq_project",
            "rna_seq_project_pi",
            "rna_seq_project_name",
            "rna_volume",
            "collection_date_inc",
        ]
        for slot in removed_slots:
            if slot in biosample:
                raise Exception(f"Field `{slot}` present in biosample {id}.")




from nmdc_schema.migrators.migrator_base import MigratorBase
from nmdc_schema.migrators.helpers import load_yaml_asset

class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "11.8.0.part_1"
    _to_version = "11.8.0.part_2"

    def upgrade(self) -> None:
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""
        self.adapter.process_each_document("data_object_set", [self.clarify_ft_data_category])

    def clarify_ft_data_category(self, data_object: dict) -> dict:
        r"""
        If the data object does not have data_category field, add a value based on the data_object_type.
        If the data_category field is already present, do nothing.
        A mapping is loaded from a YAML file. The mapping is a dictionary where the keys are data_object_type
        and the values are data_category.
        If the data_object_type is not in the mapping, raise a ValueError.

        >>> m = Migrator()
        >>> m.clarify_ft_data_category({"id": 123, "type": "nmdc:DataObject", "data_object_type": "Metagenome Raw Reads"})
        {'id': 123, 'type': 'nmdc:DataObject', 'data_object_type': 'Metagenome Raw Reads', 'data_category': 'instrument_data'}
        >>> m.clarify_ft_data_category({"id": 123, "type": "nmdc:DataObject", "data_category": "instrument_data"})
        {'id': 123, 'type': 'nmdc:DataObject', 'data_category': 'instrument_data'}
        >>> m.clarify_ft_data_category({"id": 123, "type": "nmdc:DataObject", "data_object_type": "Metag Raw"})
        Traceback (most recent call last):
        ...
        ValueError: data_object_type Metag Raw is not found in the mapping. Cannot assign data_category.
        """
        data_category_map = load_yaml_asset('migrator_from_11_7_to_11_8/data_category_map.yaml')

        if not data_object.get("data_category"):
            try:
                data_object["data_category"] = data_category_map[data_object.get("data_object_type")]
            except KeyError:
                raise ValueError(
                    f"data_object_type {data_object.get('data_object_type')} is not found in the mapping. Cannot assign data_category."
                )
        return data_object


#!/usr/bin/env python3
"""
Script to run a specific migrator against a MongoDB database using PyMongo.

Usage:
    python run_migrator.py migrator_from_11_9_1_to_11_10_0 [commit|rollback] [--host localhost] [--port 27022] [--database nmdc]
    
Arguments:
    migrator_name - Name of the migrator to run
    action - Either 'commit' to save changes or 'rollback' to discard them (default: rollback)
    
Environment variables (loaded from .env file if present):
    MONGO_HOST - MongoDB host (default: localhost)
    MONGO_PORT - MongoDB port (default: 27022)
    MONGO_DB - MongoDB database name (default: nmdc)
    MONGO_USERNAME - MongoDB username
    MONGO_PASSWORD - MongoDB password
    MONGO_AUTH_DB - Authentication database name
"""
import importlib

import click
import sys
import os
from dotenv import load_dotenv
from pymongo import MongoClient
from nmdc_schema.migrators.adapters.mongo_adapter import MongoAdapter

# TODO: Is this script somewhat redundant with nmdc_schema/migration_recursion.py? Can they be
#       consolidated?

# Load environment variables from .env file if it exists
load_dotenv()


@click.command()
@click.argument("migrator", required=True)
@click.argument("action", required=False, default="rollback", type=click.Choice(["commit", "rollback"]))
@click.option("--host", default=lambda: os.getenv("MONGO_HOST", "localhost"), help="MongoDB host (default: from MONGO_HOST env var or localhost)")
@click.option("--port", default=lambda: int(os.getenv("MONGO_PORT", "27022")), type=int, help="MongoDB port (default: from MONGO_PORT env var or 27022)")
@click.option("--database", default=lambda: os.getenv("MONGO_DB", "nmdc"), help="MongoDB database name (default: from MONGO_DB env var or nmdc)")
@click.option("--auth-db", default=lambda: os.getenv("MONGO_AUTH_DB"), help="Authentication database name (default: from MONGO_AUTH_DB env var)")
@click.option("--username", default=lambda: os.getenv("MONGO_USERNAME"), help="MongoDB username (default: from MONGO_USERNAME env var)")
@click.option("--password", default=lambda: os.getenv("MONGO_PASSWORD"), help="MongoDB password (default: from MONGO_PASSWORD env var)")
def run_migrator(migrator, action, host, port, database, auth_db, username, password):
    """Run a specific migrator against MongoDB using PyMongo.
    
    MIGRATOR should the fully qualified module name for module containing a Migrator class
    ACTION should be either 'commit' to save changes or 'rollback' to discard changes (default: rollback).
    """
    
    # Construct MongoDB connection string for PyMongo
    if username and password:
        connection_string = f"mongodb://{username}:{password}@{host}:{port}/"
        params = []
        if auth_db:
            params.append(f"authSource={auth_db}")
        # Add replica set and direct connection for Docker environments
        params.append("replicaSet=rs0")
        params.append("directConnection=true")
        if params:
            connection_string += "?" + "&".join(params)
    else:
        connection_string = f"mongodb://{host}:{port}/?replicaSet=rs0&directConnection=true"
    
    commit_changes = action == "commit"
    
    click.echo(f"Running migrator: {migrator}")
    click.echo(f"Action: {action.upper()} (changes will {'be saved' if commit_changes else 'be discarded'})")
    click.echo(f"MongoDB connection: {host}:{port}/{database}")
    
    # Get the migrator class
    migrator_module = importlib.import_module(migrator)
    migrator_class = getattr(migrator_module, "Migrator")
    
    # Connect to MongoDB using PyMongo
    try:
        client = MongoClient(connection_string)
        # Test the connection
        client.admin.command('ping')
        db = client[database]
    except Exception as e:
        click.echo(f"Error: Could not connect to MongoDB: {e}", err=True)
        sys.exit(1)
    
    # Create adapter and migrator
    try:
        adapter = MongoAdapter(database=db)
        migrator = migrator_class(adapter=adapter)
        
        # Run the migration
        migrator.upgrade(commit_changes=commit_changes)
        click.echo("Migration completed successfully!")
        
    except Exception as e:
        click.echo(f"Migration failed: {e}", err=True)
        raise
    finally:
        if 'client' in locals():
            client.close()


if __name__ == "__main__":
    run_migrator()


from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "11.6.1"
    _to_version = "11.7.0"

    def upgrade(self) -> None:
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""

        self.adapter.process_each_document("data_object_set", [self.clarify_ft_data_object_type])

    def clarify_ft_data_object_type(self, data_object: dict) -> dict:
        r"""
        If the data object has the data_object_type of "FT ICR-MS Analysis Results", replace it with "Direct Infusion FT-ICR MS Analysis Results".
        If the data object has any other value for data_object_type; ignore it.

        >>> m = Migrator()
        >>> m.clarify_ft_data_object_type({"id":123, "type":"nmdc:DataObject", "data_object_type": "FT ICR-MS Analysis Results"})
        {'id': 123, 'type': 'nmdc:DataObject', 'data_object_type': 'Direct Infusion FT-ICR MS Analysis Results'}
        >>> m.clarify_ft_data_object_type({"id":123, "type":"nmdc:DataObject", "data_object_type": "Another Type"})
        {'id': 123, 'type': 'nmdc:DataObject', 'data_object_type': 'Another Type'}
        >>> m.clarify_ft_data_object_type({"id":123, "type":"nmdc:DataObject"})
        {'id': 123, 'type': 'nmdc:DataObject'}
        """

        if data_object.get("data_object_type") == "FT ICR-MS Analysis Results":
            data_object["data_object_type"] = "Direct Infusion FT-ICR MS Analysis Results"
        return data_object


# Note: This migrator was originally introduced by @kheal in PR https://github.com/microbiomedata/nmdc-schema/pull/2276
#       It has since been renamed and relocated from its original name and location.

from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "11.2.0.part_1"
    _to_version = "11.2.0.part_2"

    def upgrade(self):
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""

        self.adapter.process_each_document("workflow_execution_set", [self.set_metap_analysis_category])

    def set_metap_analysis_category(self, workflow: dict) -> dict:
        r"""
        If the workflow execution records is of the type "nmdc:MetaproteomicsAnalysis",
        add field `metaproteomics_analysis_category` and assign it the value "matched_metagenome".

        >>> m = Migrator()
        >>> m.set_metap_analysis_category({'id': 123, 'type': 'nmdc:MetaproteomicsAnalysis'})  # field doesn't exist yet
        {'id': 123, 'type': 'nmdc:MetaproteomicsAnalysis', 'metaproteomics_analysis_category': 'matched_metagenome'}
        >>> m.set_metap_analysis_category({'id': 123, 'type': 'nmdc:MetabolomicsAnalysis'})  # not a metaproteomics analysis
        {'id': 123, 'type': 'nmdc:MetabolomicsAnalysis'}
        """

        if workflow["type"] == "nmdc:MetaproteomicsAnalysis":
            if "metaproteomics_analysis_category" not in workflow:
                workflow["metaproteomics_analysis_category"] = "matched_metagenome"
        return workflow



import logging
import re
import importlib
import inspect
import shutil

import click
import click_log
import yaml
from linkml_runtime import SchemaView

from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter

logger = logging.getLogger(__name__)
click_log.basic_config(logger)

# TODO: Is this script somewhat redundant with nmdc_schema/migrators/cli/run_migrator.py? Can they
#       be consolidated?

# todo: log before and after states of migration


def load_yaml_file(filename):
    """Loads a YAML file into a Python dict."""
    with open(filename, "r") as f:
        data = yaml.safe_load(f)
    return data


@click.command()
@click_log.simple_verbosity_option(logger)
@click.option("--schema-path", default='nmdc_schema/nmdc_materialized_patterns.yaml', required=True, type=str,
              help="Path to the schema file to which the input YAML data file conforms")
@click.option("--input-path", default='local/mongo_as_unvalidated_nmdc_database.yaml', required=True, type=str,
              help="Path to the input YAML data file")
@click.option("--output-path", default='local/rdf_safe.yaml', required=True, type=str,
              help="Path to the output YAML data file")
@click.option("--migrator-name", type=str, multiple=True,
              help="The name of a migrator module that you want to use to migrate data. If you specify this option "
                   "multiple times, the specified migrator modules will be used in the order specified.\n\n"
                   "Migrator modules are Python modules residing in the `nmdc_schema/migrators/` directory. "
                   "Their names begin with `migrator_from_`.")
def main(schema_path, input_path, output_path, migrator_name):
    """
    Generates a data file that conforms to a different schema version than the input data file does.

    See source code for initial and final schema versions.
    """

    migrators = []

    for current_name in migrator_name:
        try:
            # Import the module whose name matches the specified migrator module name.
            # Reference: https://docs.python.org/3/library/importlib.html#importlib.import_module
            migrator_module = importlib.import_module(f".{current_name}", package="nmdc_schema.migrators")

            # Get the `Migrator` class defined in that module.
            # Reference: https://docs.python.org/3/library/inspect.html#inspect.isclass
            migrator_class = getattr(migrator_module, "Migrator")
            if not inspect.isclass(migrator_class):
                raise TypeError

            # Append that class to the list of migrator classes.
            migrators.append(migrator_class)
            logger.info(f"Will use migrator: {current_name}")
        except (KeyError, ImportError, AttributeError, TypeError) as exception:
            logger.error(f"ERROR: {current_name} is not a valid migrator module name. {exception}.")
            continue

    if len(migrators) == 0:
        logger.info("No valid migrators specified. Just copying the input file to the output file.")
        # Copy the input file to the output file
        shutil.copy(input_path, output_path)
        exit()

    # Load the input data
    logger.info(f"Loading data from {input_path}")
    total_dict = load_yaml_file(input_path)

    # Load the schema and determine which of its slots we can migrate.
    logger.info(f"Loading schema from {schema_path}")
    view = SchemaView(schema_path)
    slots = view.all_slots()
    curie_slots = set()
    for sk, sv in slots.items():
        s_range = sv.range
        if s_range == "uriorcurie":
            slot_descendants = view.slot_descendants(sk)
            for i in slot_descendants:
                curie_slots.add(i)
        elif s_range:
            range_type_name = type(view.get_element(s_range)).class_name
            if range_type_name == "class_definition":
                if view.get_identifier_slot(s_range):
                    slot_descendants = view.slot_descendants(sk)
                    for i in slot_descendants:
                        curie_slots.add(i)

    logger.info(curie_slots)

    # Instantiate an adapter that the migrator can use to manipulate a "database" represented as a Python dictionary.
    # Also, specify some callback functions, so we can react when certain events occur.

    dictionary_adapter = DictionaryAdapter(
        database=total_dict,
        on_collection_created=lambda name: logger.info(f'Created collection "{name}"'),
        on_collection_renamed=lambda old_name, name: logger.info(f'Renamed collection "{old_name}" to "{name}"'),
        on_collection_deleted=lambda name: logger.info(f'Deleted collection "{name}"'),
    )

    # Iterate over the specified Migrator classes:
    for current_migrator in migrators:
        # Instantiate this migrator, binding it to the adapter we instantiated earlier.
        migrator = current_migrator(adapter=dictionary_adapter, logger=logger)
        # Invoke this migrator's `upgrade` method; effectively "doing" the migration.
        migrator.upgrade()

    # all migrations complete. save data.
    logger.info(f"Saving migrated data to {output_path}")
    with open(output_path, "w") as f:
        yaml.dump(total_dict, f)


if __name__ == "__main__":
    main()



from nmdc_schema.migrators.migrator_base import MigratorBase

class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "11.7.0"
    _to_version = "11.8.0.part_1"

    def upgrade(self) -> None:
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""
        self.adapter.do_for_each_document("data_object_set", self.validate_data_object_type)

    def validate_data_object_type(self, data_object: dict) -> None:
        r"""
        Raises an exception if the document lacks a `data_object_type` field.

        >>> m = Migrator()
        >>> m.validate_data_object_type({"id": 123, "type": "nmdc:DataObject", "data_object_type": "Type"})
        >>> m.validate_data_object_type({"id": 123, "type": "nmdc:DataObject"})
        Traceback (most recent call last):
        ...
        ValueError: data_object_type is required and is not present in the data object 123
        """
        if "data_object_type" not in data_object:
            raise ValueError(f"data_object_type is required and is not present in the data object {data_object.get('id')}")



from nmdc_schema.migrators.migrator_base import MigratorBase

class Migrator(MigratorBase):
    """
    Migrates data from X to an unknown PR, namely removes the 'designated_class' slot from
    all classes it is used.
    """

    _from_version = "X"
    _to_version = "unknown"

    def upgrade(self):
        r"""
        Migrates the database from conforming to the original schema, to conforming to the new schema.
        """

        collection_names = [
            "omics_processing_set",
            "pooling_set",
            "library_preparation_set",
            "extraction_set"
        ]
        
        for collection_name in collection_names:
            self.adapter.process_each_document(
                collection_name=collection_name,
                pipeline=[self.remove_designated_class_slot],
            )

    def remove_designated_class_slot(self, document:dict):
        r"""
        Removes the slot `designated_class` from collections that use it. 

        >>> m = Migrator()
        >>> m.remove_designated_class_slot({'id': 123, 'designated_class': 'nmdc:OmicsProcessing'}) 
        {'id': 123}
        """

        if "designated_class" in document:
            document.pop('designated_class')
            self.logger.info(f"Removing slot designated_class from doc {document['id']}")
                
        return document